@@ -33,6 +33,7 @@
 #endif
 
 #include <assert.h> /* assert() */
+#include <ctype.h> /* isdigit() isspace() */
 #include <stddef.h> /* NULL size_t */
 #include <stdlib.h> /* free() */
 #include <stdio.h> /* snprintf() */
@@ -60,6 +61,7 @@
 #include "ui/color_scheme.h"
 #include "ui/colors.h"
 #include "ui/statusbar.h"
+#include "ui/tabs.h"
 #include "utils/env.h"
 #include "utils/fs.h"
 #include "utils/macros.h"
@@ -73,6 +75,7 @@
 #include "filelist.h"
 #include "filetype.h"
 #include "plugins.h"
+#include "sort.h"
 #include "tags.h"
 
 /* State information for making completion. */
@@ -108,10 +111,13 @@ static void complete_filetype(const char *str);
 static void complete_progs(const char *str, assoc_records_t records);
 static void complete_plugin(const char str[], char *argv[], int arg_num);
 static int complete_highlight(const char args[], const char arg[], int arg_num);
-static void complete_highlight_groups(const char str[], int file_hi_only);
+static void complete_highlight_groups(const char str[], int for_clear);
 static int complete_highlight_arg(const char *str);
 static void complete_envvar(const char str[]);
 static int complete_select(completion_data_t *data);
+static void complete_wingo(const char str[]);
+static int wingo_sorter(const char a[], const char b[]);
+static const char * skip_number(const char str[]);
 static void complete_winrun(const char str[]);
 static void complete_from_string_list(const char str[], const char *items[][2],
 		size_t item_count, int ignore_case);
@@ -195,7 +201,7 @@ non_path_completion(completion_data_t *data)
 		vle_abbr_complete(args);
 		data->start = args;
 	}
-	else if(command_accepts_expr(id))
+	else if(cmds_has_expr_args(id))
 	{
 		complete_expr(arg, &data->start);
 	}
@@ -242,6 +248,11 @@ non_path_completion(completion_data_t *data)
 	}
 	else if(id == COM_KEEPSEL || id == COM_WINDO)
 		;
+	else if(id == COM_WINGO)
+	{
+		complete_wingo(args);
+		data->start = args;
+	}
 	else if(id == COM_WINRUN)
 	{
 		if(argc == 0)
@@ -260,11 +271,11 @@ non_path_completion(completion_data_t *data)
 	}
 	else if(id == COM_BMARKS && (!emark || earg_num(argc, args) >= 2))
 	{
-		bmarks_complete(argc, argv, arg);
+		bmarks_complete(MAX(0, argc - 1), argv, arg);
 	}
 	else if(id == COM_DELBMARKS && !emark)
 	{
-		bmarks_complete(argc, argv, arg);
+		bmarks_complete(MAX(0, argc - 1), argv, arg);
 	}
 	else if(id == COM_COMPARE)
 	{
@@ -504,7 +515,8 @@ complete_expr(const char str[], const char **start)
 
 		vle_opts_complete_real(*start, scope);
 	}
-	else if(dollar == NULL && !starts_with_lit(str, "v:"))
+	else if(dollar == NULL && !starts_with_lit(str, "v:") &&
+			!starts_with_lit(str, "g:"))
 	{
 		function_complete_name(str, start);
 	}
@@ -519,21 +531,29 @@ static void
 complete_compare(const char str[])
 {
 	static const char *lines[][2] = {
-		{ "byname",     "compare by file name" },
-		{ "bysize",     "compare by file size" },
-		{ "bycontents", "compare by file size and hash" },
+		{ "byname",          "compare by file name" },
+		{ "bysize",          "compare by file size" },
+		{ "bycontents",      "compare by file size and hash" },
 
-		{ "ofboth",     "use files of two views" },
-		{ "ofone",      "use files of two current view only" },
+		{ "ofboth",          "use files of two views" },
+		{ "ofone",           "use files of two current view only" },
 
-		{ "listall",    "list all files" },
-		{ "listunique", "list only unique files" },
-		{ "listdups",   "list only duplicated files" },
+		{ "listall",         "list all files" },
+		{ "listunique",      "list only unique files" },
+		{ "listdups",        "list only duplicated files" },
 
-		{ "groupids",   "group files in two panes by ids" },
-		{ "grouppaths", "group files in two panes by paths" },
+		{ "groupids",        "group files in two panes by ids" },
+		{ "grouppaths",      "group files in two panes by paths" },
 
-		{ "skipempty",  "exclude empty files from comparison" },
+		{ "skipempty",       "exclude empty files from comparison" },
+
+		{ "showidentical",   "control visibility of identical files" },
+		{ "showdifferent",   "control visibility of changed files" },
+		{ "showuniqueleft",  "control visibility of unique top/left files" },
+		{ "showuniqueright", "control visibility of unique bottom/right files" },
+
+		{ "withicase",       "force ignoring case on comparing names" },
+		{ "withrcase",       "force respecting case on comparing names" },
 	};
 
 	complete_from_string_list(str, lines, ARRAY_LEN(lines), 0);
@@ -654,6 +674,7 @@ complete_history(const char str[])
 		{ "filter",  "local filter patterns" },
 
 		{ "exprreg", "expression register input" },
+		{ "mcmd",    "command-line commands of menus" },
 	};
 	complete_from_string_list(str, lines, ARRAY_LEN(lines), 0);
 }
@@ -733,7 +754,7 @@ complete_progs(const char *str, assoc_records_t records)
 
 		if(strnoscmp(cmd, str, len) == 0)
 		{
-			vle_compl_put_match(escape_chars(cmd, "|"), assoc->description);
+			vle_compl_put_match(escape_chars(cmd, "|\\"), assoc->description);
 		}
 	}
 }
@@ -782,12 +803,49 @@ complete_highlight(const char args[], const char arg[], int arg_num)
 
 /* Completes highlight groups and subcommand "clear" for :highlight command. */
 static void
-complete_highlight_groups(const char str[], int file_hi_only)
+complete_highlight_groups(const char str[], int for_clear)
 {
 	int i;
 	const size_t len = strlen(str);
-
 	col_scheme_t *const cs = curr_stats.cs;
+
+	for(i = 0; i < SK_TOTAL; ++i)
+	{
+		if(sort_enum[i][0] == '\0' ||
+				(for_clear && cs_get_column_hi(cs, i) == NULL))
+		{
+			continue;
+		}
+
+		char name[16];
+		snprintf(name, sizeof(name), "column:%s", sort_enum[i]);
+		if(strncmp(str, name, len) == 0)
+		{
+			vle_compl_add_match(name, "");
+		}
+	}
+
+	int next_lua_id = vlua_viewcolumns_next_id(curr_stats.vlua);
+	for(i = SK_TOTAL; i < next_lua_id; ++i)
+	{
+		if(for_clear && cs_get_column_hi(cs, i) == NULL)
+		{
+			continue;
+		}
+
+		char *lua_name = vlua_viewcolumn_map_back(curr_stats.vlua, i);
+		if(lua_name != NULL)
+		{
+			char name[16];
+			snprintf(name, sizeof(name), "column:%s", lua_name);
+			if(strncmp(str, name, len) == 0)
+			{
+				vle_compl_add_match(name, "");
+			}
+			free(lua_name);
+		}
+	}
+
 	for(i = 0; i < cs->file_hi_count; ++i)
 	{
 		const char *const expr = matchers_get_expr(cs->file_hi[i].matchers);
@@ -797,7 +855,7 @@ complete_highlight_groups(const char str[], int file_hi_only)
 		}
 	}
 
-	if(!file_hi_only)
+	if(!for_clear)
 	{
 		for(i = 0; i < MAXNUM_COLOR; ++i)
 		{
@@ -904,7 +962,9 @@ complete_highlight_arg(const char *str)
 			{
 				if(strncasecmp(equal, XTERM256_COLOR_NAMES[i], len) == 0)
 				{
-					vle_compl_add_match(XTERM256_COLOR_NAMES[i], "");
+					char as_number[8];
+					snprintf(as_number, sizeof(as_number), "%d", (int)i);
+					vle_compl_add_match(XTERM256_COLOR_NAMES[i], as_number);
 				}
 			}
 
@@ -958,8 +1018,8 @@ complete_select(completion_data_t *data)
 		return data->start - args;
 	}
 
-	/* Fake !-command completion by hiding "!" in front and calling this
-		* function again. */
+	/* Fake !-command completion by hiding "!" in front and calling this function
+	 * again. */
 	cmd_info_t exec_info = *data->cmd_info;
 	char *exec_argv[exec_info.argc];
 
@@ -982,6 +1042,63 @@ complete_select(completion_data_t *data)
 	return 1 + complete_args(COM_EXECUTE, &exec_info, arg_pos, data->extra_arg);
 }
 
+/* Completes views by matching input string against their paths and titles. */
+static void
+complete_wingo(const char str[])
+{
+	vle_compl_set_sorter(&wingo_sorter);
+
+	const char *needle = skip_number(str);
+
+	int i;
+	tab_info_t tab_info;
+	for(i = 0; tabs_enum_all(i, &tab_info); ++i)
+	{
+		const view_t *view = tab_info.view;
+
+		char *title = ui_view_make_title(view);
+		if(title == NULL)
+		{
+			continue;
+		}
+
+		/* Compare case sensitive strings even on Windows. */
+		if(strstr(title, needle) != NULL ||
+				strstr(flist_get_dir(view), needle) != NULL)
+		{
+			vle_compl_put_match(format_str("%2u %s", view->id, title), "");
+		}
+		free(title);
+	}
+
+	vle_compl_finish_group();
+	vle_compl_add_last_match(str);
+}
+
+/* qsort()-like sorter for completion items of the :wingo command.  It skips
+ * leading view id so that views are sorted by their title. */
+static int
+wingo_sorter(const char a[], const char b[])
+{
+	/* Compare case sensitive strings even on Windows. */
+	return strcmp(skip_number(a), skip_number(b));
+}
+
+/* Advances over a number followed by whitespace at the beginning of a string if
+ * there is one.  The whitespace (before and after) is dropped.  Returns
+ * possibly updated input pointer. */
+static const char *
+skip_number(const char str[])
+{
+	const char *s = skip_whitespace(str);
+	while(isdigit(*s))
+	{
+		++s;
+	}
+
+	return skip_whitespace(isspace(s[0]) ? s + 1 : str);
+}
+
 /* Completes first :winrun argument. */
 static void
 complete_winrun(const char str[])
@@ -1071,6 +1188,39 @@ fast_run_complete(const char cmd[])
 	return result;
 }
 
+int
+complete_to_view_id(const char str[], unsigned int *id)
+{
+	int i;
+	tab_info_t tab_info;
+	int matches = 0;
+	for(i = 0; tabs_enum_all(i, &tab_info); ++i)
+	{
+		const view_t *view = tab_info.view;
+
+		char *title = ui_view_make_title(view);
+		if(title == NULL)
+		{
+			continue;
+		}
+
+		/* Compare case sensitive strings even on Windows. */
+		if(strstr(title, str) != NULL || strstr(flist_get_dir(view), str) != NULL)
+		{
+			if(matches == 0)
+			{
+				*id = view->id;
+			}
+
+			++matches;
+		}
+
+		free(title);
+	}
+
+	return matches;
+}
+
 /* Fills list of completions with executables in $PATH. */
 static void
 complete_command_name(const char beginning[])
@@ -1161,7 +1311,7 @@ filename_completion(const char str[], CompletionType type,
 	temp = strrchr(dirname, '/');
 	if(temp != NULL && type != CT_FILE && type != CT_FILE_WOE)
 	{
-		strcpy(filename, ++temp);
+		replace_string(&filename, ++temp);
 		*temp = '\0';
 	}
 	else if(replace_string(&dirname, ".") != 0)
@@ -1382,7 +1532,7 @@ complete_with_shared(const char *server, const char *file)
 				strcat(buf, "/");
 				if(file_matches(buf, file, len))
 				{
-					vle_compl_put_match(shell_like_escape(buf, 1), "");
+					vle_compl_add_path_match(buf);
 				}
 				p++;
 			}
@@ -1414,50 +1564,5 @@ file_matches(const char fname[], const char prefix[], size_t prefix_len)
 	return (cmp == 0);
 }
 
-int
-external_command_exists(const char cmd[])
-{
-	if(vlua_handler_cmd(curr_stats.vlua, cmd))
-	{
-		return vlua_handler_present(curr_stats.vlua, cmd);
-	}
-
-	if(curr_stats.shell_type == ST_CMD || curr_stats.shell_type == ST_YORI)
-	{
-		/* This is a builtin command. */
-		if(stroscmp(cmd, "start") == 0)
-		{
-			return 1;
-		}
-	}
-
-	char path[PATH_MAX + 1];
-	if(get_cmd_path(cmd, sizeof(path), path) == 0)
-	{
-		return executable_exists(path);
-	}
-
-	return 0;
-}
-
-int
-get_cmd_path(const char cmd[], size_t path_len, char path[])
-{
-	if(starts_with(cmd, "!!"))
-	{
-		cmd += 2;
-	}
-
-	if(contains_slash(cmd))
-	{
-		char *const expanded = replace_tilde(expand_envvars(cmd, EEF_NONE));
-		copy_str(path, path_len, expanded);
-		free(expanded);
-		return 0;
-	}
-
-	return find_cmd_in_path(cmd, path_len, path);
-}
-
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
