@@ -148,6 +148,9 @@ typedef struct
 }
 sbuffer;
 
+static parsing_result_t parse_from(const char input[],
+		expr_t (*production)(parse_context_t *ctx, const char **in), int strict,
+		int interactive);
 static int eval_expr(parse_context_t *ctx, expr_t *expr);
 static int eval_or_op(parse_context_t *ctx, int nops, expr_t ops[],
 		var_t *result);
@@ -175,7 +178,7 @@ static var_t parse_doubly_quoted_string(parse_context_t *ctx, const char **in);
 static int parse_doubly_quoted_char(parse_context_t *ctx, const char **in,
 		sbuffer *sbuf);
 static var_t eval_envvar(parse_context_t *ctx, const char **in);
-static var_t eval_builtinvar(parse_context_t *ctx, const char **in);
+static var_t eval_var(parse_context_t *ctx, const char **in);
 static var_t eval_opt(parse_context_t *ctx, const char **in);
 static expr_t parse_logical_not(parse_context_t *ctx, const char **in);
 static int parse_sequence(parse_context_t *ctx, const char **in,
@@ -203,6 +206,26 @@ vle_parser_init(getenv_func getenv_f)
 
 parsing_result_t
 vle_parser_eval(const char input[], int interactive)
+{
+	return parse_from(input, &parse_or_expr, /*strict=*/0, interactive);
+}
+
+parsing_result_t
+vle_parser_eval_call(const char input[])
+{
+	/* Unlike in Vim, don't execute call expression followed by trailing
+	 * characters. */
+	return parse_from(input, &parse_funccall, /*strict=*/1, /*interactive=*/1);
+}
+
+/* Performs parsing and evaluation.  Accepts top-level production.  Non-strict
+ * parsing means evaluation of an expression followed by trailing characters.
+ * Returns structure describing the outcome.  Field value of the result should
+ * be freed by the caller. */
+static
+parsing_result_t parse_from(const char input[],
+		expr_t (*production)(parse_context_t *ctx, const char **in),
+		int strict, int interactive)
 {
 	assert(initialized && "Parser must be initialized before use.");
 
@@ -217,7 +240,7 @@ vle_parser_eval(const char input[], int interactive)
 	};
 
 	get_next(&ctx, &ctx.last_position);
-	expr_t expr_root = parse_or_expr(&ctx, &ctx.last_position);
+	expr_t expr_root = production(&ctx, &ctx.last_position);
 	result.last_parsed_char = ctx.last_position;
 
 	result.value = var_error();
@@ -235,11 +258,15 @@ vle_parser_eval(const char input[], int interactive)
 				/* This is a comment, just ignore it. */
 				ctx.last_position += strlen(ctx.last_position);
 			}
-			else if(eval_expr(&ctx, &expr_root) == 0)
+			else if(!strict && eval_expr(&ctx, &expr_root) == 0)
 			{
 				result.value = var_clone(expr_root.value);
 				ctx.last_error = PE_INVALID_EXPRESSION;
 			}
+			else if(strict)
+			{
+				ctx.last_error = PE_INVALID_EXPRESSION;
+			}
 		}
 	}
 
@@ -322,9 +349,9 @@ eval_or_op(parse_context_t *ctx, int nops, expr_t ops[], var_t *result)
 		return 0;
 	}
 
-	/* Conversion to integer so that strings are converted into numbers instead of
-	 * checked to be empty. */
-	val = var_to_int(ops[0].value);
+	/* TODO: replace with var_to_bool() when it's OK to change semantics of
+	 *       strings by themselves. */
+	val = (var_to_int(ops[0].value) != 0);
 
 	for(i = 1; i < nops && !val; ++i)
 	{
@@ -332,7 +359,9 @@ eval_or_op(parse_context_t *ctx, int nops, expr_t ops[], var_t *result)
 		{
 			return 1;
 		}
-		val |= var_to_int(ops[i].value);
+		/* TODO: replace with var_to_bool() when it's OK to change semantics of
+		 *       strings by themselves. */
+		val |= (var_to_int(ops[i].value) != 0);
 	}
 
 	*result = var_from_bool(val);
@@ -364,9 +393,9 @@ eval_and_op(parse_context_t *ctx, int nops, expr_t ops[], var_t *result)
 		return 0;
 	}
 
-	/* Conversion to integer so that strings are converted into numbers instead of
-	 * checked to be empty. */
-	val = var_to_int(ops[0].value);
+	/* TODO: replace with var_to_bool() when it's OK to change semantics of
+	 *       strings by themselves. */
+	val = (var_to_int(ops[0].value) != 0);
 
 	for(i = 1; i < nops && val; ++i)
 	{
@@ -374,7 +403,9 @@ eval_and_op(parse_context_t *ctx, int nops, expr_t ops[], var_t *result)
 		{
 			return 1;
 		}
-		val &= var_to_int(ops[i].value);
+		/* TODO: replace with var_to_bool() when it's OK to change semantics of
+		 *       strings by themselves. */
+		val &= (var_to_int(ops[i].value) != 0);
 	}
 
 	*result = var_from_bool(val);
@@ -818,8 +849,8 @@ parse_concat_expr(parse_context_t *ctx, const char **in)
 	return result;
 }
 
-/* term ::= signed_number | number | sqstr | dqstr | envvar | builtinvar |
- *          funccall | opt | logical_not | '(' or_expr ')' */
+/* term ::= signed_number | number | sqstr | dqstr | envvar | var | funccall |
+ *          opt | logical_not | '(' or_expr ')' */
 static expr_t
 parse_term(parse_context_t *ctx, const char **in)
 {
@@ -878,7 +909,7 @@ parse_term(parse_context_t *ctx, const char **in)
 			{
 				if(**in == ':')
 				{
-					result.value = eval_builtinvar(ctx, in);
+					result.value = eval_var(ctx, in);
 				}
 				else
 				{
@@ -1101,20 +1132,23 @@ eval_envvar(parse_context_t *ctx, const char **in)
 	return var_from_str(getenv_fu(name));
 }
 
-/* builtinvar ::= 'v:' varname */
+/* var ::= 'g:' varname | 'v:' varname */
 static var_t
-eval_builtinvar(parse_context_t *ctx, const char **in)
+eval_var(parse_context_t *ctx, const char **in)
 {
 	var_t var_value;
-	char name[VAR_NAME_LENGTH_MAX + 1];
-	strcpy(name, "v:");
 
-	if(ctx->last_token.c != 'v' || **in != ':')
+	if(!ONE_OF(ctx->last_token.c, 'v', 'g') || **in != ':')
 	{
 		ctx->last_error = PE_INVALID_EXPRESSION;
 		return var_false();
 	}
 
+	char name[VAR_NAME_LENGTH_MAX + 1];
+	name[0] = ctx->last_token.c;
+	name[1] = ':';
+	name[2] = '\0';
+
 	get_next(ctx, in);
 	get_next(ctx, in);
 
@@ -1303,7 +1337,9 @@ parse_funccall(parse_context_t *ctx, const char **in)
 	{
 		ctx->last_error = PE_INVALID_EXPRESSION;
 	}
+
 	get_next(ctx, in);
+	skip_whitespace_tokens(ctx, in);
 
 	return result;
 }
